# __main__.py

from typing import Optional, Dict
import argparse

import manyworlds as mw


def main():
    parser = argparse.ArgumentParser(prog="manyworlds")
    parser.add_argument("--input", "-i", help="input scenario file")
    parser.add_argument("--output", "-o", help="output scenario file")
    parser.add_argument(
        "--mode",
        "-m",
        choices=["strict", "relaxed"],
        default="strict",
        help="flattening mode",
    )
    parser.add_argument(
        "--write-comments",
        "-c",
        default=False,
        action="store_true",
        help="output comments",
    )
    args = parser.parse_args()

    # read hierarchical feature file:
    feature = mw.Feature.from_file(args.input)

    print_feature_outline(feature)

    # write flat feature file:
    if args.output:
        feature.flatten(args.output, mode=args.mode, write_comments=args.write_comments)


def print_feature_outline(feature: mw.Feature) -> None:
    """print feature outline to terminal"""
    level_open: Dict[int, bool] = {}
    branch_shape: str
    later_sibling: bool
    sib: mw.scenario.Scenario
    sib_index: Optional[int]

    for sc in feature.scenarios():
        scenario_string = sc.name

        # Indentation and branch shapes:
        level: Optional[int] = sc.level()
        index: Optional[int] = sc.index()
        if level is None or index is None:
            continue
        if level > 1:
            # indentation:
            indentation: str = ""
            for lvl in range(2, level):
                indentation += (
                    "│   " if (lvl in level_open.keys() and level_open[lvl]) else "    "
                )

            # branch shape:
            later_sibling = False
            for sib in sc.siblings():
                sib_index = sib.index()
                if sib_index is None:
                    continue
                if sib_index > index:
                    later_sibling = True

            if later_sibling is True:
                branch_shape = "├──"
                level_open[level] = True
            else:
                branch_shape = "└──"
                level_open[level] = False

            scenario_string = indentation + branch_shape + " " + scenario_string

        # Colon for organizational scenarios:
        if sc.is_organizational():
            scenario_string += ":"

        print(scenario_string)


if __name__ == "__main__":
    main()



"""Defines the Feature Class"""

import re
import igraph as ig  # type: ignore
from typing import Optional, TextIO, Literal, List, Tuple

from .scenario import Scenario
from .step import Step, Prerequisite, Action, Assertion
from .data_table import DataTable, DataTableRow
from .exceptions import InvalidFeatureFileError


class Feature:
    """A collection of one or more directed trees
    the vertices of which represent BDD scenarios."""

    TAB_SIZE: int = 4
    """
    int

    The number of spaces per indentation level
    """

    FEATURE_PATTERN: re.Pattern = re.compile(
        r"""
        ^                    # start of line
        Feature:             # "Feature:" keyword
        [ ]                  # space
        (?P<feature_name>.*) # feature name
        $                    # end of line
        """,
        re.VERBOSE,
    )
    """
    re.Pattern

    Pattern describing a BDD feature line ("Feature: …")
    """

    COMMENT_PATTERN: re.Pattern = re.compile(
        r"""
        ^                    # start of line
        \#                   # "#" character
        [ ]                  # space
        (?P<comment>.*)      # comment
        $                    # end of line
        """,
        re.VERBOSE,
    )
    """
    re.Pattern

    Pattern describing a comment line ("# …")
    """

    graph: ig.Graph
    """The graph representing the scenario tree(s)"""
    name: Optional[str]
    """The name of the feature"""
    description: List[str]
    """The description lines for the feature"""

    def __init__(self) -> None:
        """Constructor method"""

        self.graph = ig.Graph(directed=True)
        self.name = None
        self.description = []

    @classmethod
    def split_line(cls, raw_line: str) -> Tuple[int, str]:
        """Splits a raw feature file line into the indentation part and the line part.

        Parameters
        ----------
        raw_line : str
            The raw feature file line including indentation and newline

        Returns
        -------
        tuple[int, str]
            The indentation part and the line part (without newline) as a tuple
        """

        line: str = raw_line.rstrip()
        line_wo_indentation: str = line.lstrip()
        indentation: int = len(line) - len(line_wo_indentation)
        return (indentation, line_wo_indentation)

    def parse_step_line(self, line: str) -> Optional[Step]:
        """Parses a feature file step line into the appropriate
        Step subclass instance.

        If the line begins with "And" then the step type is determined
        by the type of the last step.

        Parameters
        ----------
        line : str
            The step line (without indentation and newline)

        Returns
        -------
        Prerequisite or Action or Assertion
            An instance of a Step subclass
        """

        match: Optional[re.Match] = Step.STEP_PATTERN.match(line)
        if match is None:
            return None

        conjunction, name, comment = match.group("conjunction", "name", "comment")

        if conjunction in ["And", "But"]:
            previous_step = self.scenarios()[-1].steps[-1]
            conjunction = previous_step.conjunction

        if conjunction == "Given":
            return Prerequisite(name, comment=comment)
        elif conjunction == "When":
            return Action(name, comment=comment)
        else:  # conjunction == "Then"
            return Assertion(name, comment=comment)

    @classmethod
    def from_file(cls, file_path) -> "Feature":
        """Parses an indented feature file into a Feature instance.

        Parameters
        ----------
        file_path : str
            The path to the feature file

        Returns
        -------
        Feature
            A new Feature instance
        """

        feature = Feature()
        with open(file_path) as indented_file:
            for line_no, raw_line in enumerate(indented_file.readlines()):
                if raw_line.strip() == "":
                    continue  # Skip empty lines

                indentation: int
                line: str
                indentation, line = cls.split_line(raw_line)

                # (1) Determine and validate indentation level:
                if indentation % cls.TAB_SIZE == 0:
                    level: int = int(indentation / cls.TAB_SIZE) + 1
                else:
                    raise InvalidFeatureFileError(
                        "Invalid indentation at line {line_no}: {line}".format(
                            line_no=line_no + 1, line=line
                        )
                    )

                # (2) Parse line:

                # Feature line?
                feature_match: Optional[re.Match] = cls.FEATURE_PATTERN.match(line)
                if feature_match is not None:
                    if len(feature.scenarios()) == 0:
                        feature.name = feature_match["feature_name"]
                        continue
                    else:
                        raise InvalidFeatureFileError(
                            "Feature line is allowed only at beginning of file "
                            "but was encountered at line {line_no}: {line}".format(
                                line_no=line_no + 1, line=line
                            )
                        )

                # Scenario line?
                scenario_match: Optional[re.Match] = Scenario.SCENARIO_PATTERN.match(
                    line
                )
                if scenario_match is not None:
                    feature.append_scenario(
                        scenario_match.group("scenario_name"),
                        comment=scenario_match.group("comment"),
                        at_level=level,
                        line_no=line_no,
                    )
                    continue

                # Step line?
                new_step: Optional[Step] = feature.parse_step_line(line)
                if new_step:
                    feature.append_step(new_step, at_level=level, line_no=line_no)
                    continue

                # Data table line?
                new_data_row: Optional[DataTableRow] = DataTable.parse_line(line)
                if new_data_row:
                    feature.append_data_row(
                        new_data_row, at_level=level, line_no=line_no
                    )
                    continue

                # Comment line?
                comment_match: Optional[re.Match] = cls.COMMENT_PATTERN.match(line)
                if comment_match is not None:
                    continue  # skip comment lines

                # Feature description line?
                if feature.name is not None and len(feature.scenarios()) == 0:
                    feature.description.append(line)
                    continue

                # Not a valid line!
                raise InvalidFeatureFileError(
                    "Unable to parse line {line_no}: {line}".format(
                        line_no=line_no + 1, line=line
                    )
                )

        return feature

    def append_scenario(
        self, scenario_name: str, comment: Optional[str], at_level: int, line_no: int
    ) -> Scenario:
        """Append a scenario to the feature.

        Parameters
        ----------
        scenario : Scenario
            The scenario to append

        comment : str, optional
            A comment

        at_level : int
            The indentation level of the scenario in the input file.
            Used for indentation validation.

        line_no : int
            The line number of the scenario in the input file.
            Used in InvalidFeatureFile error message.
        """

        if at_level > 1:  # Non-root scenario:
            # Find the parent to connect scenario to:
            parent_level: int = at_level - 1
            parent_level_scenarios: List[Scenario] = [
                sc
                for sc in self.scenarios()
                if sc.level() == parent_level and not sc.is_closed()
            ]
            if len(parent_level_scenarios) > 0:
                return Scenario(
                    scenario_name,
                    self.graph,
                    parent_scenario=parent_level_scenarios[-1],
                    comment=comment,
                )
            else:
                raise InvalidFeatureFileError(
                    "Excessive indentation at line {line_no}: Scenario: {name}".format(
                        line_no=line_no + 1, name=scenario_name
                    )
                )

        else:  # Root scenario:
            return Scenario(scenario_name, self.graph, comment=comment)

    def append_step(self, step: Step, at_level: int, line_no: int) -> None:
        """Appends a step to the feature.

        Parameters
        ----------
        step : Prerequisite or Action or Assertion
            The Step subclass instance to append

        at_level : int
            The level at which to add the step.
            Used for indentation validation.

        line_no : int
            The line number of the step in the input file.
            Used in InvalidFeatureFile error message.
        """

        # Ensure the indentation level of the step matches
        # the last scenario indentation level
        last_scenario: Scenario = self.scenarios()[-1]
        if at_level == last_scenario.level():
            last_scenario.steps.append(step)
        else:
            raise InvalidFeatureFileError(
                "Invalid indentation at line {line_no}: {name}".format(
                    line_no=line_no + 1, name=step.name
                )
            )

    def append_data_row(
        self, data_row: DataTableRow, at_level: int, line_no: int
    ) -> None:
        """Appends a data row to the feature.

        Adds a data table to the last step if necessary
        Otherwise adds row to data table.

        Parameters
        ----------
        data_row : DataTableRow
            The data row to append

        at_level : int
            The level at which to add the data row.
            Used for indentation validation.

        line_no : int
            The line number of the data row in the input file.
            Used in InvalidFeatureFile error message.
        """

        last_step: Step = self.scenarios()[-1].steps[-1]
        if last_step.data:
            # Row is an additional row for an existing table
            last_step.data.rows.append(data_row)
        else:
            # Row is the header row of a new table
            last_step.data = DataTable(data_row)

    @classmethod
    def write_feature_declaration(cls, file_handle: TextIO, feature: "Feature") -> None:
        """Writes feature name and (optional) description
        to the end of a flat feature file.

        Parameters
        ----------
        file_handle : TextIO
            The file to which to append the feature declaration
        """
        if feature.name is not None:
            file_handle.write(
                "Feature: {feature_name}\n\n".format(feature_name=feature.name)
            )
        if len(feature.description) > 0:
            for line in feature.description:
                file_handle.write("    {line}\n".format(line=line))
            file_handle.write("\n")

    @classmethod
    def write_scenario_name(
        cls, file_handle: TextIO, scenarios: List[Scenario], write_comment: bool = False
    ) -> None:
        """Writes formatted scenario name to the end of a "relaxed" flat feature file.

        Parameters
        ----------
        file_handle : TextIO
            The file to which to append the scenario name

        scenarios : List[Scenario]
            Organizational and validated scenarios along the path

        write_comment : bool, default = False
            Whether or not to write comment if present
        """

        # (1) Group consecutive regular or organizational scenarios:
        groups: List[List[Scenario]] = []

        # Function for determining whether a scenario can be added to a current group:
        def group_available_for_scenario(
            gr: List[List[Scenario]], sc: Scenario
        ) -> bool:
            return (
                len(gr) > 0
                and len(gr[-1]) > 0
                and gr[-1][-1].is_organizational() == sc.is_organizational()
            )

        for sc in scenarios:
            if group_available_for_scenario(groups, sc):
                groups[-1].append(sc)  # add to current group
            else:
                groups.append([sc])  # start new group

        # (2) Format each group to strings:
        group_strings: List[str] = []

        for group in groups:
            if group[-1].is_organizational():
                group_strings.append(
                    "[{}]".format(" / ".join([sc.name for sc in group]))
                )
            else:
                group_strings.append(" > ".join([sc.name for sc in group]))

        # (3) Assemble name:
        scenario_string: str = "Scenario: {}".format(" ".join(group_strings))

        # (4) Optional comment:
        destination_scenario: Scenario = scenarios[-1]
        if write_comment is True and destination_scenario.comment is not None:
            scenario_string += " # {comment}".format(
                comment=destination_scenario.comment
            )

        # (4) Write name:
        file_handle.write(scenario_string + "\n")

    @classmethod
    def write_scenario_steps(
        cls, file_handle: TextIO, steps: List[Step], write_comments: bool = False
    ) -> None:
        """Writes formatted scenario steps to the end of the flat feature file.

        Parameters
        ----------
        file_handle : io.TextIOWrapper
            The file to which to append the steps

        steps : List[Step]
            Steps to append to file_handle

        write_comments: bool, default = False
            Whether or not to write comments if present
        """

        last_step: Optional[Step] = None
        for step in steps:
            first_of_type: bool = (
                last_step is None or last_step.conjunction != step.conjunction
            )
            step_string: str = step.format(first_of_type=first_of_type)
            if write_comments is True and step.comment is not None:
                step_string += " # {comment}".format(comment=step.comment)
            file_handle.write(step_string + "\n")

            if step.data:
                Feature.write_data_table(
                    file_handle, step.data, write_comment=write_comments
                )
            last_step = step

    @classmethod
    def write_data_table(
        cls, file_handle: TextIO, data_table: DataTable, write_comment: bool = False
    ) -> None:
        """Writes formatted data table to the end of the flat feature file.

        Parameters
        ----------
        file_handle : io.TextIOWrapper
            The file to which to append the data table

        data_table : DataTable
            A data table

        write_comment : bool
            Whether or not to write comment if present
        """

        # Determine column widths to accommodate all values:
        col_widths: List[int] = [
            max([len(cell) for cell in col])
            for col in list(zip(*data_table.to_list_of_list()))
        ]

        for row in data_table.to_list():
            # pad values with spaces to column width:
            padded_row: List[str] = [
                row.values[col_num].ljust(col_width)
                for col_num, col_width in enumerate(col_widths)
            ]

            # add column enclosing pipes:
            table_row_string: str = "    | {columns} |".format(
                columns=" | ".join(padded_row)
            )

            # add comments:
            if write_comment is True and row.comment is not None:
                table_row_string += " # {comment}".format(comment=row.comment)

            # write line:
            file_handle.write(table_row_string + "\n")

    def flatten(
        self,
        file_path: str,
        mode: Literal["strict", "relaxed"] = "strict",
        write_comments: bool = False,
    ) -> None:
        """Writes a flat (no indentation) feature file representing the feature.

        Parameters
        ----------
        file_path : str
            Path to flat feature file to be written

        mode : {"strict", "relaxed"}, default="strict"
            Flattening mode. Either "strict" or "relaxed"

        comments : bool, default = False
            Whether or not to write comments
        """

        with open(file_path, "w") as flat_file:
            # Feature declaration:
            if self.name is not None:
                Feature.write_feature_declaration(flat_file, self)

            # Scenarios:
            if mode == "strict":
                self.flatten_strict(flat_file, write_comments=write_comments)
            elif mode == "relaxed":
                self.flatten_relaxed(flat_file, write_comments=write_comments)

    def flatten_strict(self, flat_file: TextIO, write_comments: bool = False) -> None:
        """Write. a flat (no indentation) feature file representing the feature
        using the "strict" flattening mode.

        The "strict" flattening mode writes one scenario per vertex in the tree,
        resulting in a feature file with one set of "When" steps followed by one
        set of "Then" steps (generally recommended).

        Parameters
        ----------
        flat_file : io.TextIOWrapper
            The flat feature file

        write_comments : bool, default = False
            Whether or not to write comments
        """

        for scenario in [sc for sc in self.scenarios() if not sc.is_organizational()]:
            # Scenario name:
            scenarios_for_naming: List[Scenario] = [
                sc
                for sc in scenario.path_scenarios()
                if sc.is_organizational() or sc == scenario
            ]
            Feature.write_scenario_name(
                flat_file, scenarios_for_naming, write_comment=write_comments
            )

            ancestor_scenarios = scenario.ancestors()
            steps: List[Step] = []
            # collect prerequisites from all scenarios along the path
            steps += [st for sc in ancestor_scenarios for st in sc.prerequisites()]
            # collect actions from all scenarios along the path
            steps += [st for sc in ancestor_scenarios for st in sc.actions()]
            # add all steps from the destination scenario only
            steps += scenario.steps

            # Write steps:
            Feature.write_scenario_steps(
                flat_file, steps, write_comments=write_comments
            )
            flat_file.write("\n")  # Empty line to separate scenarios

    def flatten_relaxed(self, flat_file: TextIO, write_comments: bool = False) -> None:
        """Writes a flat (no indentation) feature file representing the feature
        using the "relaxed" flattening mode.

        The "relaxed" flattening mode writes one scenario per leaf vertex in the tree,
        resulting in a feature file with multiple consecutive sets of "When" and "Then"
        steps per scenario (generally considered an anti-pattern).

        Parameters
        ----------
        flat_file : io.TextIOWrapper
            The flat feature file

        write_comments : bool, default = False
            Whether or not to write comments if present
        """

        for scenario in self.leaf_scenarios():
            steps: List[Step] = []
            # organizational and validated scenarios used for naming:
            scenarios_for_naming: List[Scenario] = []
            for path_scenario in scenario.path_scenarios():
                steps += path_scenario.prerequisites()
                steps += path_scenario.actions()
                if path_scenario.is_organizational():
                    scenarios_for_naming.append(path_scenario)
                elif not path_scenario.validated:
                    steps += path_scenario.assertions()
                    path_scenario.validated = True
                    scenarios_for_naming.append(path_scenario)

            Feature.write_scenario_name(
                flat_file, scenarios_for_naming, write_comment=write_comments
            )
            # Write steps:
            Feature.write_scenario_steps(
                flat_file, steps, write_comments=write_comments
            )
            flat_file.write("\n")  # Empty line to separate scenarios

    def find(self, *scenario_names: List[str]) -> Optional[Scenario]:
        """Finds and returns a scenario by the names of all scenarios along the path
        from a root scenario to the destination scenario.

        Used in tests only

        Parameters
        ----------
        scenario_names : List[str]
            List of scenario names

        Returns
        -------
        Scenario or None
            The found scenario, or None if none found
        """

        # Root scenario:
        scenario: Optional[Scenario] = next(
            (sc for sc in self.root_scenarios() if sc.name == scenario_names[0]), None
        )
        if scenario is None:
            return None

        # Root descendant scenarios:
        for scenario_name in scenario_names[1:]:
            scenario = next(
                (
                    vt["scenario"]
                    for vt in scenario.vertex.successors()
                    if vt["scenario"].name == scenario_name
                ),
                None,
            )
            if scenario is None:
                return None

        return scenario

    def scenarios(self) -> List[Scenario]:
        """Returns all scenarios

        Returns
        -------
        List[Scenario]
            All scenarios in index order
        """

        return [vx["scenario"] for vx in self.graph.vs]

    def root_scenarios(self) -> List[Scenario]:
        """Returns the root scenarios (scenarios with vertices without incoming edges).

        Returns
        -------
        List[Scenario]
            All root scenarios in index order
        """
        return [vx["scenario"] for vx in self.graph.vs if vx.indegree() == 0]

    def leaf_scenarios(self) -> List[Scenario]:
        """Returns the leaf scenarios (scenarios with vertices without outgoing edges).

        Returns
        -------
        List[Scenario]
            All leaf scenarios in index order
        """
        return [vx["scenario"] for vx in self.graph.vs if vx.outdegree() == 0]



"""Defines the DataTable and DataTableRow classes"""

import re
from typing import Optional, List


class DataTableRow:
    """A Gherkin data table row"""

    values: List[str]
    comment: Optional[str]

    def __init__(self, values: List[str], comment: Optional[str] = None):
        """Constructor method

        Parameters
        ----------
        values : List[str]
            The header row

        comment : str
            Comment (optional)
        """

        self.values = values
        self.comment = comment


class DataTable:
    """A Gherkin data table"""

    TABLE_ROW_PATTERN = re.compile(
        r"""
        ^                                  # start of line
        (?P<table_row>\| (?:[^|]+[ ]+\|)+) # pipe-delimited list of values
        (?:[ ]+\#[ ](?P<comment>.+))?      # optional comment
        $                                  # end of line
        """,
        re.VERBOSE,
    )
    """
    re.Pattern

    Pattern describing a Gherkin data table row
    followed by an optional comment
    """

    header_row: DataTableRow
    rows: List[DataTableRow]

    def __init__(self, header_row: DataTableRow) -> None:
        """Constructor method

        Parameters
        ----------
        header_row : DataTableRow
            The header row
        """

        self.header_row = header_row
        self.rows = []

    def to_list_of_list(self) -> List[List[str]]:
        """Returns a list of list of str representation of itself

        First row is header row

        Returns
        -------
        List[List[str]]
            The list of list of str representation of itself
        """

        return [self.header_row.values] + [row.values for row in self.rows]

    def to_list_of_dict(self) -> List[dict]:
        """Returns a list of dict representation of itself

        Returns
        -------
        List[dict]
            The list of dict representation of itself
        """

        return [dict(zip(self.header_row.values, row.values)) for row in self.rows]

    def to_list(self) -> List[DataTableRow]:
        """Returns a list of DataTableRow representation of itself

        Returns
        -------
        List[DataTableRow]
            The list of DataTableRow representation of itself
        """

        return [self.header_row] + self.rows

    @classmethod
    def parse_line(cls, line: str) -> Optional[DataTableRow]:
        """Parses a pipe delimited data table line into a DataTableRow

        Parameters
        ----------
        line : str
            A pipe delimited data table line

        Returns
        -------
        DataTableRow
        """

        match: Optional[re.Match] = DataTable.TABLE_ROW_PATTERN.match(line)

        if match:
            values = [s.strip() for s in match.group("table_row").split("|")[1:-1]]
            comment = match.group("comment")
            return DataTableRow(values, comment)
        else:
            return None



"""Test the Feature class"""

import os
import filecmp

import pytest

import manyworlds as mw


@pytest.fixture(scope="session", autouse=True)
def clear_out_directory():
    """Delete all files in test/out"""
    folder = os.path.dirname(os.path.realpath(__file__)) + "/out"
    for filename in os.listdir(folder):
        if not filename == ".gitignore":
            file_path = os.path.join(folder, filename)
            os.unlink(file_path)
    yield


def test_find():
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")

    # Finds existing root scenario:
    root_scenario = feature.find("View users")
    assert type(root_scenario) == mw.scenario.Scenario

    # Does not find anything if first name provided is not a root scenario:
    root_scenario = feature.find("Select user")
    assert root_scenario is None

    # Finds existing non-root scenario
    non_root_scenario = feature.find("View users", "Bulk operations", "Select user")
    assert non_root_scenario is not None

    # Does not find non-existing non-root scenario (path scenario missing)
    non_root_scenario = feature.find("View users", "Select user")
    assert non_root_scenario is None


def test_from_file():
    """Test the structure of the feature graph after using the 'from_file' method"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")

    assert feature.name == "User Deactivation"
    assert len(feature.description) == 3

    assert len(feature.root_scenarios()) == 1

    # root scenario:
    root_scenario = feature.find("View users")
    assert root_scenario.name == "View users"
    assert len(root_scenario.prerequisites()) == 1
    assert len(root_scenario.actions()) == 1
    assert len(root_scenario.assertions()) == 1
    data = root_scenario.prerequisites()[0].data.to_list_of_dict()
    assert len(data) == 4
    assert data[2]["Name"] == "Connie"
    assert data[2]["Status"] == "Active"

    # leaf scenario:
    leaf_scenario = feature.find(
        "View users",
        "Bulk operations",
        "Select user",
        "Select multiple users",
        "Bulk deactivate users",
        "Confirm bulk deactivation of users",
    )
    assert leaf_scenario.name == "Confirm bulk deactivation of users"
    assert len(leaf_scenario.prerequisites()) == 0
    assert len(leaf_scenario.actions()) == 1
    assert len(leaf_scenario.assertions()) == 2
    assert leaf_scenario.comment == 'by clicking "OK"'


def test_flatten_strict():
    """Test the 'flatten' method in 'strict' mode"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    feature.flatten("test/out/scenarios_flat_strict.feature")
    assert filecmp.cmp(
        "test/out/scenarios_flat_strict.feature",
        "test/fixtures/out/scenarios_flat_strict.feature",
    )


def test_flatten_strict_with_comments():
    """Test the 'flatten' method in 'strict' mode with comments turned on"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    feature.flatten(
        "test/out/scenarios_flat_strict_with_comments.feature", write_comments=True
    )
    assert filecmp.cmp(
        "test/out/scenarios_flat_strict_with_comments.feature",
        "test/fixtures/out/scenarios_flat_strict_with_comments.feature",
    )


def test_flatten_relaxed():
    """Test the 'flatten' method in 'relaxed' mode"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    feature.flatten("test/out/scenarios_flat_relaxed.feature", mode="relaxed")
    assert filecmp.cmp(
        "test/out/scenarios_flat_relaxed.feature",
        "test/fixtures/out/scenarios_flat_relaxed.feature",
    )


def test_flatten_relaxed_with_comments():
    """Test the 'flatten' method in 'relaxed' mode with comments turned on"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    feature.flatten(
        "test/out/scenarios_flat_relaxed_with_comments.feature",
        mode="relaxed",
        write_comments=True,
    )
    assert filecmp.cmp(
        "test/out/scenarios_flat_relaxed_with_comments.feature",
        "test/fixtures/out/scenarios_flat_relaxed_with_comments.feature",
    )


def test_organizational_scenarios():
    """Test the correct output of organizational scenarios"""
    feature = mw.Feature.from_file(
        "test/fixtures/in/feature_with_organizational_scenarios.feature"
    )
    feature.flatten(
        "test/out/scenarios_flat_strict_with_organizational_scenarios.feature",
        mode="strict",
    )
    assert filecmp.cmp(
        "test/out/scenarios_flat_strict_with_organizational_scenarios.feature",
        "test/fixtures/out/scenarios_flat_strict_with_organizational_scenarios.feature",
    )



"""Test the Scenario class"""

import pytest

import manyworlds as mw


@pytest.fixture(scope="module")
def root_scenario():
    """load a representative root scenario"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    return feature.find("View users")


@pytest.fixture(scope="module")
def leaf_scenario():
    """load a representative leaf scenario"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    return feature.find(
        "View users",
        "Bulk operations",
        "Select user",
        "Select multiple users",
        "Bulk deactivate users",
        "Confirm bulk deactivation of users",
    )


def test_repr(root_scenario, leaf_scenario):
    assert (
        root_scenario.__repr__() == "<Scenario: View users "
        "(1 prerequisites, 1 actions, 1 assertions)>"
    )
    assert (
        leaf_scenario.__repr__() == "<Scenario: Confirm bulk deactivation of users "
        "(0 prerequisites, 1 actions, 2 assertions)>"
    )


def test_level(root_scenario, leaf_scenario):
    assert root_scenario.level() == 1
    assert leaf_scenario.level() == 6


def test_ancestors(root_scenario, leaf_scenario):
    assert len(root_scenario.ancestors()) == 0
    assert len(leaf_scenario.ancestors()) == 5


def test_index(root_scenario, leaf_scenario):
    assert root_scenario.index() == 0
    assert leaf_scenario.index() == 8


def test_parent(root_scenario, leaf_scenario):
    assert root_scenario.parent() is None
    assert leaf_scenario.parent().name == "Bulk deactivate users"


def test_children(root_scenario, leaf_scenario):
    assert len(root_scenario.children()) == 2
    assert len(leaf_scenario.children()) == 0


def test_siblings(root_scenario, leaf_scenario):
    assert len(root_scenario.siblings()) == 1
    assert len(leaf_scenario.siblings()) == 2



"""Test the CLI"""

import os
import filecmp

import pytest


@pytest.fixture(scope="session", autouse=True)
def clear_out_directory():
    """Delete all files in test/out"""
    folder = os.path.dirname(os.path.realpath(__file__)) + "/out"
    for filename in os.listdir(folder):
        if not filename == ".gitignore":
            file_path = os.path.join(folder, filename)
            os.unlink(file_path)
    yield


def test_cli():
    exit_status = os.system("python -m manyworlds --help")
    assert exit_status == 0


def test_cli_hierarchy_output():
    exit_status = os.system(
        "python -m manyworlds --input test/fixtures/in/feature.feature "
        "--output test/out/scenarios_flat_strict_cli.feature "
        "> test/out/scenario_hierarchy.txt"
    )
    assert exit_status == 0
    assert filecmp.cmp(
        "test/out/scenario_hierarchy.txt", "test/fixtures/out/scenario_hierarchy.txt"
    )



"""Test handling of invalid input files"""

import pytest

import manyworlds as mw


def test_invalid_file_mis_spelled_conjunction():
    """Test that the correct error is raised when attempting to parse invalid files"""
    with pytest.raises(mw.exceptions.InvalidFeatureFileError) as error_info:
        mw.Feature.from_file("test/fixtures/in/invalid/invalid_conjunction.feature")
    assert (
        str(error_info.value)
        == 'Unable to parse line 2: Whenx I go to "Users" # mis-spelled conjunction'
    )


def test_invalid_file_invalid_indentation():
    """Test that the correct error is raised when attempting to parse invalid files"""
    with pytest.raises(mw.exceptions.InvalidFeatureFileError) as error_info:
        mw.Feature.from_file("test/fixtures/in/invalid/invalid_indentation.feature")
    assert (
        str(error_info.value) == "Invalid indentation at line 5: Scenario: "
        "Indented using 3 spaces instead of 4"
    )


def test_invalid_file_excessive_indentation():
    """Test that the correct error is raised when attempting to parse invalid files"""
    with pytest.raises(mw.exceptions.InvalidFeatureFileError) as error_info:
        mw.Feature.from_file("test/fixtures/in/invalid/excessive_indentation.feature")
    assert (
        str(error_info.value) == "Excessive indentation at line 13: Scenario: "
        "Indented 2 instead of 1 levels"
    )


def test_invalid_file_excessive_step_indentation():
    """Test that the correct error is raised when attempting to parse invalid files"""
    with pytest.raises(mw.exceptions.InvalidFeatureFileError) as error_info:
        mw.Feature.from_file(
            "test/fixtures/in/invalid/excessive_step_indentation.feature"
        )
    assert str(error_info.value) == "Invalid indentation at line 3: " "I see users"


def test_invalid_file_feature_after_scenario():
    """Test that the correct error is raised when attempting to parse invalid files"""
    with pytest.raises(mw.exceptions.InvalidFeatureFileError) as error_info:
        mw.Feature.from_file("test/fixtures/in/invalid/feature_after_scenario.feature")
    assert (
        str(error_info.value) == "Feature line is allowed only at beginning of file "
        "but was encountered at line 5: Feature: User Deactivation"
    )



"""Defines the Scenario Class"""
# needed to support "type[]" class type annotations in Python 3.8:
from __future__ import annotations

import re
import igraph as ig  # type: ignore
from typing import Optional, Union, List

from .step import Step, Prerequisite, Action, Assertion


class Scenario:
    """A BDD Scenario"""

    SCENARIO_PATTERN: re.Pattern = re.compile(
        r"""
        ^                        # start of line
        Scenario:                # "Scenario:" keyword
        [ ]                      # space
        (?P<scenario_name>[^#]*) # scenario name
        (?:\# (?P<comment>.+))?  # optional comment
        $                        # end of line
        """,
        re.VERBOSE,
    )
    """
    re.Pattern

    Pattern describing a BDD scenario line ("Scenario: …")
    followed by an optional comment
    """

    name: str
    graph: ig.Graph
    vertex: ig.Vertex
    steps: List[Step]
    comment: Optional[str]
    _validated: bool

    def __init__(
        self,
        name: str,
        graph: ig.Graph,
        parent_scenario: Optional["Scenario"] = None,
        comment: Optional[str] = None,
    ) -> None:
        """Constructor method

        Parameters
        ----------
        name : str
            The name of the scenario

        graph : igraph.Graph
            The graph

        parent_scenario: Scenario (optional)
            The parent scenario to connect the new scenario to

        comment : str, optional
        """

        self.name = name.strip()
        self.graph = graph
        self.vertex = graph.add_vertex()
        self.vertex["scenario"] = self
        self.steps = []
        self._validated = False
        self.comment = comment.strip() if comment is not None else None
        if parent_scenario is not None:
            self.graph.add_edge(parent_scenario.vertex, self.vertex)

    @property
    def validated(self) -> bool:
        """The "validated" property

        Used to keep track of which scenarios had their assertions written
        to an output scenario already so that assertions are not run multiple
        times.

        Returns
        -------
        bool
            Whether or not this scenario has been validated
        """
        return self._validated

    @validated.setter
    def validated(self, value: bool) -> None:
        """The validated property setter

        Parameters
        ----------
        value : bool
            Whether or not this scenario has been validated

        """

        self._validated = value

    def prerequisites(self) -> List[Step]:
        """Returns all steps of type Prerequisite

        Returns
        -------
        List[Prerequisite]
            List of steps of type Prerequisite
        """

        return self.steps_of_type(Prerequisite)

    def actions(self) -> List[Step]:
        """Returns all steps of type Action

        Returns
        -------
        List[Action]
            List of steps of type Action
        """

        return self.steps_of_type(Action)

    def assertions(self) -> List[Step]:
        """Returns all steps of type Assertion

        Returns
        ----------
        list
            List[Assertion]
                List of steps of type Assertion
        """

        return self.steps_of_type(Assertion)

    def steps_of_type(
        self, step_type: Union[type[Prerequisite], type[Action], type[Assertion]]
    ) -> List[Step]:
        """Returns all steps of the passed in type

        Parameters
        ----------
        step_class : {Prerequisite, Action, Assertion}
            A step subclass

        Returns
        -------
        List[Step]
            All steps of the passed in type
        """

        return [st for st in self.steps if type(st) is step_type]

    def __str__(self) -> str:
        """Returns a string representation of the Scenario instance for terminal output.

        Returns
        -------
        str
            String representation of the Scenario instance
        """

        return "<Scenario: {} ({} prerequisites, {} actions, {} assertions)>".format(
            self.name,
            len(self.prerequisites()),
            len(self.actions()),
            len(self.assertions()),
        )

    def __repr__(self) -> str:
        """Returns a string representation of the Scenario instance for terminal output.

        Returns
        -------
        str
            String representation of the Scenario instance
        """

        return self.__str__()

    def ancestors(self) -> List["Scenario"]:
        """Returns the scenario's ancestors, starting with a root scenario

        Returns
        -------
        List[Scenario]
            List of scenarios
        """

        ancestors: List[ig.Vertex] = self.graph.neighborhood(
            self.vertex,
            mode="IN",
            order=1000,
            mindist=1,
        )
        ancestors.reverse()
        return [vx["scenario"] for vx in self.graph.vs(ancestors)]

    def parent(self) -> Optional[Scenario]:
        """Returns the scenario's parent scenario, if one exists

        Returns
        -------
        Scenario, optional
            The parent scenario
        """

        parents: List[ig.Vertex] = self.graph.neighborhood(
            self.vertex, mode="IN", order=1, mindist=1
        )
        if len(parents) == 1:
            return self.graph.vs[parents[0]]["scenario"]
        else:
            return None

    def children(self) -> List[Scenario]:
        """Returns the scenario's child scenarios

        Returns
        -------
        List[Scenario]
            The child scenarios
        """

        children: List[ig.Vertex] = self.graph.neighborhood(
            self.vertex, mode="OUT", order=1, mindist=1
        )
        return [vx["scenario"] for vx in self.graph.vs(children)]

    def siblings(self) -> List[Scenario]:
        """Returns the scenario's sibling scenarios

        The scenario's parent's children (including self)

        Returns
        -------
        List[Scenario]
            The sibling scenarios
        """

        parent: Optional[Scenario] = self.parent()
        if parent is not None:
            return parent.children()
        else:
            return [vx["scenario"] for vx in self.graph.vs if vx.indegree() == 0]
            # TODO: This duplicates the implementation of
            # ScenarioForest#root_scenarios() but Scenario does not currently
            # have access to its feature. Might want to change that

    def path_scenarios(self) -> List["Scenario"]:
        """Returns the complete scenario path from the root scenario to
        (and including) self.

        Returns
        -------
        List[Scenario]
            List of scenarios. The last scenario is self
        """

        return self.ancestors() + [self]

    def level(self) -> int:
        """Returns the scenario"s level in the scenario tree.

        Root scenario =  Level 1

        Returns
        -------
        int
            The scenario"s level
        """

        return self.graph.neighborhood_size(self.vertex, mode="IN", order=1000)

    def is_organizational(self) -> bool:
        """Returns whether the scenario is an "organizational" scenario.

        "Organizational" scenarios are used for grouping only.
        They do not have any assertions.

        Returns
        ----------
        bool
            Whether the scenario is an "organizational" scenario
        """

        return len(self.assertions()) == 0

    def index(self) -> Optional[int]:
        """Returns the "index" of the scenario.

        The scenario"s vertical position in the feature file.

        Returns
        ----------
        int
            Index of self
        """

        return self.vertex.index  # TODO: start at index 1 (instead of 0)

    def is_closed(self) -> bool:
        """Returns whether or not the scenario is "closed".

        A scenario is "closed" if additional child scenarios cannot be added
        which is the case when there is a "later" (higher index) scenario
        with a lower or equal indentation level in the feature file.

        Returns
        ----------
        bool
            Whether or not the scenario is "closed"
        """

        # Later scenario with lower or equal indentation level:
        closing_scenario: Optional[Scenario] = next(
            (
                vx
                for vx in self.graph.vs()
                if vx.index > self.index()
                and self.graph.neighborhood_size(vx, mode="IN", order=1000)
                <= self.level()
            ),
            None,
        )
        return closing_scenario is not None



"""Test the Step class"""

import pytest

import manyworlds as mw


@pytest.fixture(scope="module")
def given_step():
    """load a representative root scenario"""
    feature = mw.Feature.from_file("test/fixtures/in/feature.feature")
    return feature.find("View users").steps[0]


def test_repr(given_step):
    assert given_step.__repr__() == "<Prerequisite: The following users:>"


def test_format(given_step):
    assert given_step.format() == "Given the following users:"
    assert given_step.format(first_of_type=False) == " And the following users:"



"""Defines the Step Class and subclasses"""

import re
from typing import Optional, Literal

from .data_table import DataTable


class Step:
    """A BDD scenario step"""

    STEP_PATTERN: re.Pattern = re.compile(
        r"""
        ^                                        # start of line
        (?P<conjunction>Given|When|Then|And|But) # conjunction
        [ ]                                      # space
        (?P<name>[^#]+)                          # name
        (?:\#[ ](?P<comment>.+))?                # optional comment
        $                                        # end of line
        """,
        re.VERBOSE,
    )
    """
    re.Pattern

    Pattern describing a BDD step line ("Given", "When", ...)
    with optional comment
    """

    name: str
    conjunction: Literal["Given", "When", "Then"]
    data: Optional[DataTable]
    comment: Optional[str]

    def __init__(
        self, name: str, data: Optional[DataTable] = None, comment: Optional[str] = None
    ) -> None:
        """Constructor method

        Parameters
        ----------
        name : str
            The name of the step

        data : DataTable, optional
            A data table

        comment : str, optional
            A comment
        """

        self.name = name.strip()
        self.data = data
        self.comment = comment

    def format(self, first_of_type: bool = True) -> str:
        """Returns a string representation of the Step instance
        for feature file output.

        Uses "And" as a conjunction if the step is not the first
        of its type.

        Parameters
        ----------
        first_of_type : bool
            Whether or not the step is the first of it"s type

        Returns
        -------
        str
            String representation of the Step instance
        """

        return "{conjunction} {name}".format(
            conjunction=self.conjunction if first_of_type else " And", name=self.name
        )

    def __str__(self) -> str:
        """Return. a string representation of the Step instance
        for terminal output.

        Returns
        -------
        str
            String representation of the Step instance
        """

        return "<{type}: {name}>".format(
            type=self.__class__.__name__, name=(self.name[0].upper() + self.name[1:])
        )

    def __repr__(self) -> str:
        """Return a string representation of the Step instance
        for terminal output.

        Returns
        -------
        str
            String representation of the Step instance
        """

        return self.__str__()


class Prerequisite(Step):
    """A BDD scenario prerequisite ("Given") step"""

    def __init__(
        self, name: str, data: Optional[DataTable] = None, comment: Optional[str] = None
    ) -> None:
        self.conjunction = "Given"
        super().__init__(name, data=data, comment=comment)


class Action(Step):
    """A BDD scenario action ("When") step"""

    def __init__(
        self, name: str, data: Optional[DataTable] = None, comment: Optional[str] = None
    ) -> None:
        self.conjunction = "When"
        super().__init__(name, data=data, comment=comment)


class Assertion(Step):
    """A BDD scenario assertion ("Then") step"""

    def __init__(
        self, name: str, data: Optional[DataTable] = None, comment: Optional[str] = None
    ) -> None:
        self.conjunction = "Then"
        super().__init__(name, data=data, comment=comment)
